# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.GameCoordinator do
  use GenServer

  @moduledoc """
  The game coordinator is responsible for keeping track of active
  instances of each game and can create new ones on demand.
  """

  def start_link(initial) when is_list(initial) do
    GenServer.start_link(__MODULE__, initial, name: {:global, __MODULE__})
  end

  @impl true
  def init(initial) do
    {:ok, initial}
  end

  @impl true
  def handle_call({:find, type}, _from, state) do
    {pid, state} = _find(type, state)
    {:reply, pid, state}
  end

  @impl true
  def handle_call({:create, type}, _from, state) do
    {pid, state} = _create(type, state)
    {:reply, pid, state}
  end

  @impl true
  def handle_call({:find_or_create, type}, from, state) do
    {pid, state} = _find(type, state)
    case pid do
      nil -> handle_call({:create, type}, from, state)
      some -> {:reply, some, state}
    end
  end

  defp _create(type, games) do
    state = type.instantiate()
    pid = spawn(Amethyst.API.Game, :start, [type, state])
    games = [{type, pid, []} | games]
    {pid, games}
  end

  defp _find(type, games) do
    alive_games = games |> Enum.filter(fn {_mod, pid, _opts} -> Process.alive?(pid) end)
    # TODO: Here we should have some additional filtering for specifically joinable games
    existing = alive_games |> Enum.filter(fn {mod, _pid, _opts} -> mod == type end)
    if length(existing) > 0 do
      [{_mod, pid, _} | _] = existing
      {pid, alive_games}
    else
      {nil, alive_games}
    end
  end

  def create(type) when is_atom(type) do
    GenServer.call({:global, __MODULE__}, {:create, type})
  end
  def find(type) when is_atom(type) do
    GenServer.call({:global, __MODULE__}, {:find, type})
  end
  def find_or_create(type) when is_atom(type) do
    GenServer.call({:global, __MODULE__}, {:find_or_create, type})
  end
end
